import os
import threading
import json
import logging
import atexit
from datetime import datetime
from typing import Dict, Any, Callable
from .apiCore import WebsocketManager

__all__ = [
    'start_api',
    'restart_api',
    'Line',
    'Bar',
    'Sequence',
    'Lines',
    'Bars',
    'Sequences',
    'Scatter',
    'Area',
    'Areas',
    'Pie',
    'Radar',
    'Surface',
]

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')

_manager: WebsocketManager | None = None

def start_config_load():
    from .configEdit import config_load
    config = config_load()

    host = config['WEBSOCKET_CONFIG']['HOST']
    port = config['WEBSOCKET_CONFIG']['PORT']
    route = config['WEBSOCKET_CONFIG']['ROUTE']

    return host, port, route

def start_config_check(host: str | None, port: str | None, route: str | None):
    if host is None or port is None or route is None:
        hostDefault, portDefault, routeDefault = start_config_load()
        host = hostDefault if host is None else host
        port = portDefault if port is None else port
        route = routeDefault if route is None else route
    
    return host, port, route

def start_manager(host: str, port: str, route: str):

    global _manager
    
    logging.info("Initializing user API.")
    _manager = WebsocketManager(host=host, port=port, route=route)
    _manager.start_server_thread()
    atexit.register(_manager.stop_server_thread)
    logging.info("User API initialized.")

    return _manager

def start_api(host: str | None=None, port: str | None=None, route: str | None=None):

    global _manager
    
    if _manager is not None:
        logging.info("Data service is already running.")
        return _manager
    else:
        host, port, route = start_config_check(host, port, route)
        return start_manager(host, port, route)

def restart_api(host: str | None=None, port: str | None=None, route: str | None=None):

    global _manager

    if _manager is not None:
        logging.info("Data service is already running, restarting...")
        _manager = None
    else:
        logging.info("Data service is not running, starting...")
    return start_api(host=host, port=port, route=route)


class DataStream:
    """
    Abstract class for data streams, defines the synchronous interface
    that users can call.
    """
    @staticmethod
    def _data_validated(data_payload: Any):
        if (not isinstance(data_payload, dict)) or ("id" not in data_payload) or ("timestamp" not in data_payload) or ("value" not in data_payload):
            return False
        else:
            return True

    @staticmethod
    def _data_validated_number(data_payload: Any):
        from numbers import Number
        if DataStream._data_validated(data_payload) and isinstance(data_payload['value'], Number):
            return True
        else:
            return False

    @staticmethod
    def _data_validated_dict(data_payload: Any):
        if DataStream._data_validated(data_payload) and isinstance(data_payload['value'], dict):
            return True
        else:
            return False

    @staticmethod
    def _data_validated_list(data_payload: Any):
        if DataStream._data_validated(data_payload) and isinstance(data_payload['value'], list):
            return True
        else:
            return False

    @staticmethod
    def _data_validated_coordinate(data_payload: Any):
        from numbers import Number
        if DataStream._data_validated_list(data_payload):
            value = data_payload['value']
            if len(value) == 2:
                x, y = value
                if isinstance(x, Number) and isinstance(y, Number):
                    return True
                else:
                    return False
            else:
                return False 
        else:
            return False

    @staticmethod
    def _data_validated_dimension(data_payload: Any):
        if DataStream._data_validated_list(data_payload):
            value = data_payload['value']
            if len(value) == 2:
                dimension, num = value
                if isinstance(dimension, list) and isinstance(num, list) and len(dimension) == len(num):
                    return True
                else:
                    return False
            return False
        else:
            return False

    @staticmethod
    def _data_validated_dimensions(data_payload: Any):
        if DataStream._data_validated_list(data_payload):
            value = data_payload['value']
            if len(value) == 3:
                dimension, series, num = value
                if isinstance(dimension, list) and isinstance(series, list)  and isinstance(num, list) and len(series) == len(num):
                    return True
                else:
                    return False
            return False
        else:
            return False

    @staticmethod
    def _data_validated_surface(data_payload: Any):
        if DataStream._data_validated_list(data_payload):
            value = data_payload['value']
            if len(value) == 3:
                axis, shape, num = value
                if len(axis) == 3 and len(shape) == 2 and len(num) == shape[0] * shape[1]:
                    return True
                else:
                    return False
            else:
                return False
        else:
            return False


    def __init__(self, key_word: str, chart_type: str):
        self.key_word = key_word.strip().lower()
        self.chart_type = chart_type
        self.data_key = self._get_data_key()
        
        # Register itself with the backend Manager so it can be identified when the frontend subscribes
        _manager.register_data_stream(self.data_key)
        logging.info(f"Registered new DataStream: {self.chart_type} -> {self.key_word}")

    def _get_data_key(self) -> str:
        """Generates a unique key for the backend Manager to identify the data stream"""
        return f"{self.chart_type}<:>{self.key_word}"

    def update(self, data_payload: Dict[str, Any]):
        """Updates the data and triggers a push"""
        # Bridge the synchronous call to the asynchronous Manager, which handles cache updates and WebSocket pushes in the background event loop
        _manager.push_update_sync(self.data_key, data_payload)
        logging.debug(f"Pushed update for {self.chart_type} -> {self.key_word}")

    def get_cached_data(self) -> Dict[str, Any] | None:
        return _manager.get_cached_data_sync(self.data_key)
    
    def execute(self, logic_func: Callable, *args, **kwargs):
        """Excute a func at background"""
        
        def thread_target(stream_instance, func, *func_args, **func_kwargs):
            try:
                func(stream_instance, *func_args, **func_kwargs)
            except Exception as e:
                print(f"Error in background thread: {e}")
            finally:
                print(f"Thread finished.")

        thread = threading.Thread(
            target=thread_target, 
            args=(self, logic_func) + args, 
            kwargs=kwargs,
            daemon=True
        )
        
        thread.start()
        logging.info(f"{self.chart_type}('{self.data_key}') calls {logic_func.__name__} in background (daemon) thread.")

    def fresh(self, data_payload_value: Any):
        data_payload = {
            "id": datetime.now().strftime("%Y%m%d%H%M%S%f"),
            "timestamp": datetime.now().isoformat(),
            "value": data_payload_value
        }
        self.update(data_payload)


class Line(DataStream):
    def __init__(self, key_word: str):
        super().__init__(key_word, chart_type='line')

    def update(self, data_payload: Dict[str, Any]):
        if not DataStream._data_validated_number(data_payload):
            logging.error(f"Invalid data update form for "+str(self.chart_type)+' -> '+str(self.key_word)+r". Must be like: {id:xxx, timestamp:xxx, value:some_number}")
        else:
            super().update(data_payload)


class Bar(DataStream):
    def __init__(self, key_word: str):
        super().__init__(key_word, chart_type='bar')

    def update(self, data_payload: Dict[str, Any]):
        if not DataStream._data_validated_number(data_payload):
            logging.error(f"Invalid data update form for "+str(self.chart_type)+' -> '+str(self.key_word)+r". Must be like: {id:xxx, timestamp:xxx, value:some_number}")
        else:
            super().update(data_payload)

class Sequence(DataStream):
    def __init__(self, key_word: str):
        super().__init__(key_word, chart_type='sequence')
    
    def update(self, data_payload: Dict[str, Any]):
        if not DataStream._data_validated_number(data_payload):
            logging.error(f"Invalid data update form for "+str(self.chart_type)+' -> '+str(self.key_word)+r". Must be like: {id:xxx, timestamp:xxx, value:some_number}")
        else:
            super().update(data_payload)

class Lines(DataStream):
    def __init__(self, key_word: str):
        super().__init__(key_word, chart_type='lines')

    def update(self, data_payload: Dict[str, Any]):
        if not DataStream._data_validated_dict(data_payload):
            logging.error("Invalid data update form for "+str(self.chart_type)+' -> '+str(self.key_word)+r". Must be like: {id:xxx, timestamp:xxx, value:{A:some_number, B:some_number}}.")
        else:
            super().update(data_payload)

class Bars(DataStream):
    def __init__(self, key_word: str):
        super().__init__(key_word, chart_type='bars')
    
    def update(self, data_payload: Dict[str, Any]):
        if not DataStream._data_validated_dict(data_payload):
            logging.error("Invalid data update form for "+str(self.chart_type)+' -> '+str(self.key_word)+r". Must be like: {id:xxx, timestamp:xxx, value:{A:some_number, B:some_number}}.")
        else:
            super().update(data_payload)

class Sequences(DataStream):
    def __init__(self, key_word: str):
        super().__init__(key_word, chart_type='sequences')
    
    def update(self, data_payload: Dict[str, Any]):
        if not DataStream._data_validated_dict(data_payload):
            logging.error("Invalid data update form for "+str(self.chart_type)+' -> '+str(self.key_word)+r". Must be like: {id:xxx, timestamp:xxx, value:{A:some_number, B:some_number}}.")
        else:
            super().update(data_payload)

class Scatter(DataStream):
    def __init__(self, key_word: str):
        super().__init__(key_word, chart_type='scatter')

    def update(self, data_payload: Dict[str, Any]):
        if not DataStream._data_validated_coordinate(data_payload):
            logging.error("Invalid data update form for "+str(self.chart_type)+' -> '+str(self.key_word)+r". Must be like: {id:xxx, timestamp:xxx, value:[some_number, some_number]}.")
        else:
            super().update(data_payload)

class Area(DataStream):
    def __init__(self, key_word: str):
        super().__init__(key_word, chart_type='area')

    def update(self, data_payload: Dict[str, Any]):
        """
        Data form should be like: {id:xxx, timestamp:xxx, value:[[A, B, C], [1, 2, 3]]}
        First element of value will be used as x-axis tickers,
        Second element of value will be a list of numbers, which are the true value at different x-axis tickers.
        """
        if not DataStream._data_validated_dimension(data_payload):
            logging.error("Invalid data update form for "+str(self.chart_type)+' -> '+str(self.key_word)+r". Must be like: {id:xxx, timestamp:xxx, value:[[A, B, C], [1, 2, 3]]}.")
        else:
            super().update(data_payload)

class Areas(DataStream):
    def __init__(self, key_word: str):
        super().__init__(key_word, chart_type='areas')

    def update(self, data_payload: Dict[str, Any]):
        """
        Data form should be like: {id:xxx, timestamp:xxx, value:[[A, B, C], [label_1, label_2], [[1, 2, 3],[4, 5, 6]]]}
        First element of value will be used as x-axis tickers,
        Second element of value will be label of different data series, 
        Third element of value will be a 2-dimension array, which is the true value of different data series.
        """
        if not DataStream._data_validated_dimensions(data_payload):
            logging.error("Invalid data update form for "+str(self.chart_type)+' -> '+str(self.key_word)+r". Must be like: {id:xxx, timestamp:xxx, value:[[A, B, C], [label_1, label_2], [[1, 2, 3],[4, 5, 6]]]}.")
        else:
            super().update(data_payload)

class Pie(DataStream):
    def __init__(self, key_word: str):
        super().__init__(key_word, chart_type='pie')
    
    def update(self, data_payload: Dict[str, Any]):
        """
        Data form should be like: {id:xxx, timestamp:xxx, value:[[A, B, C], [1, 2, 3]]}
        First element of value will be used as x-axis tickers,
        Second element of value will be a list of numbers, which is the true value at different x-axis tickers.
        """
        if not DataStream._data_validated_dimension(data_payload):
            logging.error("Invalid data update form for "+str(self.chart_type)+' -> '+str(self.key_word)+r". Must be like: {id:xxx, timestamp:xxx, value:[[A, B, C], [1, 2, 3]]}.")
        else:
            super().update(data_payload)

class Radar(DataStream):
    def __init__(self, key_word: str):
        super().__init__(key_word, chart_type='radar')

    def update(self, data_payload: Dict[str, Any]):
        """
        Data form should be like: {id:xxx, timestamp:xxx, value:[[A, B, C], [100, 100, 100], [4, 5, 6]]}
        First element of value will be used as x-axis tickers,
        Second element of value will be max value at different x-axis tickers, 
        Third element of value will be a list of number, which is the true value at different x-axis tickers.
        """
        if not DataStream._data_validated_dimensions(data_payload):
            logging.error("Invalid data update form for "+str(self.chart_type)+' -> '+str(self.key_word)+r". Must be like: {id:xxx, timestamp:xxx, value:[[A, B, C], [100, 100, 100], [4, 5, 6]]}.")
        else:
            super().update(data_payload)

class Surface(DataStream):
    def __init__(self, key_word: str):
        super().__init__(key_word, chart_type='surface')
    
    def update(self, data_payload: Dict[str, Any]):
        """
        Data form should be like: {id:xxx, timestamp:xxx, value:[[A, B, C], [1, 2], [[1.2, 2.2, 9],[3.2, 4.3, 8]]]}
        First element of value will be [x-axis name, y-axis name, z-axis name],
        Second element of value will be shape of value array, which is [number of rows, number of columns], 
        Third element of value will be a list of coordinates, which is like [[x1, y1, z1], [x2, y2, z2]...].
        """
        if not DataStream._data_validated_surface(data_payload):
            logging.error("Invalid data update form for "+str(self.chart_type)+' -> '+str(self.key_word)+r". Must be like: {id:xxx, timestamp:xxx, value:[[A, B, C], [1, 2], [[1.2, 2.2, 9],[3.2, 4.3, 8]]]}.")
        else:
            super().update(data_payload)





import os
import json

path_current = os.path.dirname(os.path.abspath(__file__))
path_config = os.path.join(path_current, 'config.json' )
path_config_default = os.path.join(path_current, 'configDefault.json' )

def config_load_from(path: str):
    with open(path, 'r', encoding='utf-8') as f:
        content = json.load(f)
    return content

def config_save_to(content: dict, path: str):
    with open(path, 'w', encoding='utf-8') as f:
        json.dump(content, f, ensure_ascii=False, indent=4)

def config_load():
    return config_load_from(path_config)

def config_save(content: dict):
    config_save_to(content, path_config)

def config_reset():
    default = config_load_from(path_config_default)
    config_save(default)

def config_update_by(content: dict, config_type: str, config_item: str, config_value: str | int):
    content[config_type][config_item] = config_value
    return content

def config_update(config_type: str, config_item: str, config_value: str | int):
    content = config_load()
    content = config_update_by(content, config_type, config_item, config_value)
    config_save(content)







import sys,os
import asyncio
import threading
import json
import logging
from typing import Dict, Any, Set, Deque, Callable, Awaitable
from collections import deque
from websockets.server import serve, WebSocketServerProtocol
from websockets.exceptions import ConnectionClosedOK, ConnectionClosedError

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

# Asynchronous wrapper to handle synchronous data retrieval in user code, returning initial/cached data on frontend subscription
async def _simulate_initial_data_fetch(data_key: str, cache_data: Dict[str, Any], all_valid_keys: Set[str]):
    """Simulates data check and retrieval upon frontend subscription"""
    if data_key not in all_valid_keys:
        logger.error(f'Subscribed key does not exist. Subscribed as {data_key}, Validated keys are {all_valid_keys}')
        return None # Subscribed key does not exist
        
    if data_key in cache_data:
        # If cached, return the cache directly
        return cache_data[data_key]
    else:
        # Simulate generating an empty or default initial data structure
        chart_type, key_word = data_key.split('<:>', 1)
        return {
            "id": "INIT",
            "timestamp": "N/A",
            "value": 0
        }


class WebsocketManager:
    """Backend manager that runs the WebSocket server and asyncio event loop in a separate thread. It acts as a bridge between synchronous and asynchronous code."""
    def __init__(self, host: str, port: int, route: str):
        
        # Connection info
        self.host = host
        self.port = port
        self.route = route

        # Asynchronous core components
        self.loop: asyncio.AbstractEventLoop = None
        self._server_task: asyncio.Task = None
        
        # Thread components
        self._server_thread: threading.Thread = None
        self._is_running = False
        
        # Data and connection management, accessed in the async thread, requires protection 
        self._cache: Dict[str, Dict[str, Any]] = {} 
        self._subscriptions: Dict[str, Set[WebSocketServerProtocol]] = {}
        self._valid_data_keys: Set[str] = set() # Records all valid keys registered by the user via Line('test')
        
        # Sync/Async communication queue, used to bridge synchronous calls to the asynchronous loop
        self._update_queue: Deque[tuple[str, Dict[str, Any]]] = deque()
        self._update_event: asyncio.Event = asyncio.Event() # Notifies the async loop that new data is available

        # Thread Lock
        self._lock = threading.Lock()

    # Synchronous call bridge functions for the user-facing API

    def register_data_stream(self, data_key: str):
        """Called by the user to register a data stream"""
        with self._lock:
            self._valid_data_keys.add(data_key)

    def push_update_sync(self, data_key: str, data_payload: Dict[str, Any]):
        """Called by the user, puts data into the queue"""
        self._update_queue.append((data_key, data_payload))
        if self.loop and self.loop.is_running():
             # Wakes up the background thread to process data
            self.loop.call_soon_threadsafe(self._update_event.set)
    
    def get_cached_data_sync(self, data_key: str) -> Dict[str, Any] | None:
        """Reads the cache"""
        with self._lock:
            return self._cache.get(data_key)
        
    # Thread management

    def start_server_thread(self):
        """Starts the background thread to run the WebSocket server"""
        if self._is_running:
            return
        self._is_running = True
        self._server_thread = threading.Thread(target=self._run_in_thread, daemon=True)
        self._server_thread.start()
        logger.info("WebsocketManager started in background thread.")

    def _run_in_thread(self):
        """Thread entry function, sets up the asyncio event loop and runs the server"""
        try:
            self.loop = asyncio.new_event_loop()
            asyncio.set_event_loop(self.loop)
            
            # Start the WebSocket server and data processing task
            self.loop.create_task(self._start_server_and_loop())
            self.loop.run_forever()
        except KeyboardInterrupt:
            logger.info("Background server thread interrupted.")
        except Exception as e:
            logger.error(f"Background server thread error: {e}")
        finally:
            self.loop.stop()
            self._is_running = False

    def stop_server_thread(self):
        if not self._is_running:
            return

        self._is_running = False
        self._update_event.set()
        
        if self.loop and self.loop.is_running():
            future = asyncio.run_coroutine_threadsafe(self._shutdown_async(), self.loop)
            try:
                future.result(timeout=5)
            except:
                logger.error("Asynchronous shutdown timed out or failed.")
        
        if self._server_thread and self._server_thread.is_alive():
            self._server_thread.join(timeout=2)
            if self._server_thread.is_alive():
                 logger.warning("Background thread did not terminate gracefully.")

    async def _shutdown_async(self):

        close_tasks = []
        # Collect all alive websocket
        active_websockets = set()
        for subscriptions in self._subscriptions.values():
            active_websockets.update(subscriptions)

        for websocket in active_websockets:
            # Send close order gracefully (code=1000)
            close_tasks.append(websocket.close(code=1000, reason="Server shutting down"))
        
        # Close
        if close_tasks:
            # Make sure all action is finished
            await asyncio.gather(*close_tasks, return_exceptions=True)

        # Cancell server
        if self._server_task:
            self._server_task.cancel()
            
        # Cancel tasks
        tasks = [t for t in asyncio.all_tasks(self.loop) if t is not self._server_task and t is not asyncio.current_task()]
        for task in tasks:
            task.cancel()
        
        # Stop event loop
        self.loop.call_soon_threadsafe(self.loop.stop)

    async def _start_server_and_loop(self):
            """Starts the WebSocket server and data processing coroutine in the asynchronous loop"""
            
            # Start the WebSocket server
            try:
                # serve() returns an awaitable, convert this awaitable into a task
                server = await serve(
                    self._websocket_handler, 
                    self.host, 
                    self.port, 
                    subprotocols=["json"]
                )
                server_coroutine = server.serve_forever() # Get a coroutine task that runs indefinitely
                self._server_task = self.loop.create_task(server_coroutine)
                logger.info(f"WebSocket Server running on ws://{self.host}:{self.port}")
            except Exception as e:
                logger.error(f"Failed to start WebSocket server: {e}")
                return # Exit if startup fails
            # Start the synchronous data queue processing coroutine
            self.loop.create_task(self._process_update_queue())

    # --- Asynchronous Core Logic ---

    async def _process_update_queue(self):
        """Periodically or event-driven checks the synchronous queue and handles data push"""
        while self._is_running:
            # Use wait_for to wait for event or timeout
            await self._update_event.wait()
            self._update_event.clear()
            
            # Check for new data
            while self._update_queue:
                try:
                    data_key, new_data = self._update_queue.popleft()
                    # Execute update and push in the asynchronous loop
                    await self._update_and_push_async(data_key, new_data)
                except Exception as e:
                    logger.error(f"Error processing update queue: {e}")


    async def _update_and_push_async(self, data_key: str, new_data: Dict[str, Any]):
        """Asynchronously updates the cache and pushes to subscribers"""
        
        # Update cache
        self._cache[data_key] = new_data
        
        # Get the set of subscribers
        subscribers = self._subscriptions.get(data_key, set())
        
        if not subscribers:
            logger.debug(f"Data updated for {data_key}, but no active subscribers.")
            return

        # Asynchronous parallel push
        message = json.dumps(new_data)
        tasks = []
        disconnected_websockets = set()
        
        for websocket in subscribers:
            tasks.append(self._safe_send(websocket, message, disconnected_websockets))
        
        await asyncio.gather(*tasks)

        # Clean up disconnected connections
        if disconnected_websockets:
            self._subscriptions[data_key] -= disconnected_websockets
            logger.info(f"Cleaned up {len(disconnected_websockets)} disconnected clients for {data_key}.")
            if not self._subscriptions[data_key]:
                del self._subscriptions[data_key]
                logger.info(f"No subscribers left for {data_key}. Cleaning up subscription entry.")

    # --- WebSocket Protocol Handling ---

    async def _websocket_handler(self, websocket: WebSocketServerProtocol, path: str):
        """Handles new WebSocket connections and subscriptions"""
        client_address = f"{websocket.remote_address[0]}:{websocket.remote_address[1]}"
        data_key = "" # Used to clean up subscription upon disconnection

        if self.route and path != self.route:
            logger.warning(f"Invalid subscription request from '{client_address}'. Connection attempt on invalid path: {path}. Expected: {self.route}")
            await websocket.send(json.dumps({"status": "failure", "message": "Connection attempt on invalid path."}))
            await websocket.close(code=1008, reason="Connection attempt on invalid path.")
            return

        try:
            # Receive subscription message (expecting the frontend to send {"chart_type": "...", "key_word": "..."})
            subscription_message_raw = await websocket.recv()
            subscription_message = json.loads(subscription_message_raw)
            
            chart_type = subscription_message.get("chart_type")
            key_word = subscription_message.get("key_word")
            data_key = f"{chart_type}<:>{key_word}"

            # Check if the data stream has been registered by the user
            initial_data = await _simulate_initial_data_fetch(data_key, self._cache, self._valid_data_keys)
            
            if initial_data is None:
                logger.warning(f"Invalid subscription request from '{client_address}': chart_type='{chart_type}', key_word='{key_word}'.")
                await websocket.send(json.dumps({"status": "failure", "message": "Invalid chart type or keyword (not registered)."}))
                await websocket.close(code=1008, reason="Invalid subscription format.")
                return

            # Register subscription, add WebSocket to the set
            if data_key not in self._subscriptions:
                self._subscriptions[data_key] = set()
                
            self._subscriptions[data_key].add(websocket)
            logger.info(f"Client '{client_address}' subscribing to: chart_type='{chart_type}', key_word='{key_word}'. Total subscription: {len(self._subscriptions[data_key])}")

            # Successful response, and push initial/cached data
            await websocket.send(json.dumps({"status": "success", "message": "Subscription successful."}))
            # Push cached data
            await websocket.send(json.dumps(initial_data))
            
            # Keep the connection open, waiting for disconnection or message reception, e.g., unsubscribe message
            async for message in websocket:
                # Can handle unsubscribe or other control commands here
                logger.debug(f"Received message from {client_address}: {message}")
                pass
            
        # Catch Exception
        except ConnectionClosedOK:
            logger.info(f"Client disconnected gracefully: {client_address}")
        except asyncio.CancelledError: 
            logger.info(f"Connection task cancelled for: {client_address}")
        except ConnectionClosedError as e:
            # Client broken with exception
            logger.info(f"Client disconnected unexpectedly: {client_address} - {e}")
        except json.JSONDecodeError:
            logger.error(f"Received malformed JSON from client: {client_address}")
        except Exception as e:
            logger.error(f"Error handling connection with {client_address}: {type(e).__name__} - {e}")
        finally:
            # Connection disconnected, remove subscription
            if data_key and websocket in self._subscriptions.get(data_key, set()):
                self._subscriptions[data_key].remove(websocket)
                logger.info(f"Client {client_address} unsubscribed from {data_key}. Remaining: {len(self._subscriptions[data_key])}")


    async def _safe_send(self, websocket: WebSocketServerProtocol, message: str, disconnected_websockets: Set[WebSocketServerProtocol]):
        """Safely sends a message, catching connection closed exceptions"""
        try:
            await websocket.send(message) 
        except Exception as e:
            logger.error(f"Unexpected error during WebSocket send: {e}", exc_info=True)
            disconnected_websockets.add(websocket)